/***************************************************************************
 *   (C) 2007 by Tomas Mandys <tomas.mandys at 2p dot cz>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation;             *
 *                                                                         *
 ***************************************************************************/


/* 
 *Shared library add-on to iptables to add RTPPROXY target support.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <iptables.h>
#include <net/if.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stddef.h>
//#include <linux/netfilter_ipv4/ip_tables.h>
#include <linux/netfilter/xt_RTPPROXY.h>
#include <unistd.h>

/* Function which prints out usage message. */
static void
help(void) {
	printf(
"RTPPROXY target v%s options:\n"
"  --addr-a   ip                 Address of RTP proxy gate A\n"
"  --addr-b   port               Start port number of gate A\n"
"  --port-a   ip                 Address of RTP proxy gate B\n"
"  --port-b   port               Start port number of gate B\n"
"  --max-sess num                Max number of sessions\n"
"  --expires-timeout sec         Expiration timeout\n"
"  --resurrection-timeout sec    Resurrection timeout\n"
"\n",
IPTABLES_VERSION);
}

static struct option opts[] = {
	{ .name = "addr-a",               .has_arg = 1, .flag = 0, .val = '1' },
	{ .name = "addr-b",               .has_arg = 1, .flag = 0, .val = '2' },
	{ .name = "port-a",               .has_arg = 1, .flag = 0, .val = '3' },
	{ .name = "port-b",               .has_arg = 1, .flag = 0, .val = '4' },
	{ .name = "max-sess",             .has_arg = 1, .flag = 0, .val = '5' },
	{ .name = "expires-timeout",      .has_arg = 1, .flag = 0, .val = '6' },
	{ .name = "resurrection-timeout", .has_arg = 1, .flag = 0, .val = '7' },
	{ .name = 0 }
};

#define DEFAULT_EXPIRES_TIMEOUT 30
#define DEFAULT_RESURRECTION_TIMEOUT 40

/* Initialize the target. */
static void
init(struct ipt_entry_target *t, unsigned int *nfcache) {
	struct xt_rtpproxy_target_info *info = 
		(struct xt_rtpproxy_target_info*)t->data;

	memset(info, 0, sizeof(*info));
	info->expires_timeout = DEFAULT_EXPIRES_TIMEOUT;
	info->resurrection_timeout = DEFAULT_RESURRECTION_TIMEOUT;
}

#define IPT_RTPPROXY_OPT_ADDR_A      0x01
#define IPT_RTPPROXY_OPT_ADDR_B      0x02
#define IPT_RTPPROXY_OPT_PORT_A      0x04
#define IPT_RTPPROXY_OPT_PORT_B      0x08
#define IPT_RTPPROXY_OPT_MAX_SESS    0x10
#define IPT_RTPPROXY_OPT_EXPIRES_TIMEOUT      0x20
#define IPT_RTPPROXY_OPT_RESURRECTION_TIMEOUT 0x40

#define ADJUST_U_INT16(x) ((x)>0xFFFF?0xFFFF:(x))

/* Function which parses command options; returns true if it
   ate an option */
static int
parse(int c, char **argv, int invert, unsigned int *flags,
      const struct ipt_entry *entry,
      struct ipt_entry_target **target) {
	struct xt_rtpproxy_target_info *info = 
		(struct xt_rtpproxy_target_info*)(*target)->data;
	switch (c) {
	case '1':
		if (*flags & IPT_RTPPROXY_OPT_ADDR_A)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --addr-a twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --addr-a");

		if (!inet_aton(optarg, (struct in_addr*) &info->gate[0].ip) ) {
			exit_error(PARAMETER_PROBLEM,
				   "Invalid IP address %s",
				   optarg);
		}
  
		*flags |= IPT_RTPPROXY_OPT_ADDR_A;
		break;

	case '2':
		if (*flags & IPT_RTPPROXY_OPT_ADDR_B)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --addr-b twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --addr-b");

		if (!inet_aton(optarg, (struct in_addr*)&info->gate[1].ip)) {
			exit_error(PARAMETER_PROBLEM,
				   "Invalid IP address %s",
				   optarg);
		}

		*flags |= IPT_RTPPROXY_OPT_ADDR_B;
		break;

	case '3':
		if (*flags & IPT_RTPPROXY_OPT_PORT_A)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --port-a twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --port-a");

		if ( (info->gate[0].port = ADJUST_U_INT16(atol(optarg))) == 0)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);

		*flags |= IPT_RTPPROXY_OPT_PORT_A;
		break;

	case '4':
		if (*flags & IPT_RTPPROXY_OPT_PORT_B)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --port-b twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --port-b");

		if ( (info->gate[1].port = ADJUST_U_INT16(atol(optarg))) == 0)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);

		*flags |= IPT_RTPPROXY_OPT_PORT_B;
		break;

	case '5':
		if (*flags & IPT_RTPPROXY_OPT_MAX_SESS)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --max-sess twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --max-sess");

		if ( (info->max_sess = ADJUST_U_INT16(atol(optarg))) == 0)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);

		*flags |= IPT_RTPPROXY_OPT_MAX_SESS;
		break;

	case '6':
		if (*flags & IPT_RTPPROXY_OPT_EXPIRES_TIMEOUT)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --expires-timeout twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --expires-timeout");

		if ( (info->expires_timeout = ADJUST_U_INT16(atol(optarg))) == 0)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);

		*flags |= IPT_RTPPROXY_OPT_EXPIRES_TIMEOUT;
		break;

	case '7':
		if (*flags & IPT_RTPPROXY_OPT_RESURRECTION_TIMEOUT)
			exit_error(PARAMETER_PROBLEM,
				   "Can't specify --resurrection-timeout twice");

		if (check_inverse(optarg, &invert, NULL, 0))
			exit_error(PARAMETER_PROBLEM,
				   "Unexpected `!' after --resurrection-timeout");

		if ( (info->resurrection_timeout = ADJUST_U_INT16(atol(optarg))) == 0)
			exit_error(PARAMETER_PROBLEM,
				   "Invalid number %s",
				   optarg);

		*flags |= IPT_RTPPROXY_OPT_RESURRECTION_TIMEOUT;
		break;

	default:
		return 0;
	}
	info->libipt_flags = *flags;
	/* We must try to calc implicit params always when possible because final_check, the best place to do it, won't get info as param. Such harakiri is necessary because one could try to find entry in userspace (via userspacesize) */
	if ((*flags & (IPT_RTPPROXY_OPT_ADDR_A|IPT_RTPPROXY_OPT_PORT_A|IPT_RTPPROXY_OPT_MAX_SESS)) == (IPT_RTPPROXY_OPT_ADDR_A|IPT_RTPPROXY_OPT_PORT_A|IPT_RTPPROXY_OPT_MAX_SESS)) {
		/* we have all mandatory params */

		if ((*flags & IPT_RTPPROXY_OPT_ADDR_B) == 0) {
			info->gate[1].ip = info->gate[0].ip;
		}
		if ((*flags & IPT_RTPPROXY_OPT_PORT_B) == 0) {
			if (info->gate[1].ip == info->gate[0].ip) {
				int p;
				p = info->gate[0].port + 2*info->max_sess;
				info->gate[1].port = p<=0xFFFFL?p:0;
			}
			else
				info->gate[1].port = info->gate[0].port;
		}
	}
	return 1;
}


static void
final_check(unsigned int flags) {
	/* Grrr... we don't know struct ipt_entry_target* so we cannot make aggregated test here but in ipt_checkentry :-( */
	if ((flags & (IPT_RTPPROXY_OPT_ADDR_A|IPT_RTPPROXY_OPT_PORT_A|IPT_RTPPROXY_OPT_MAX_SESS)) != (IPT_RTPPROXY_OPT_ADDR_A|IPT_RTPPROXY_OPT_PORT_A|IPT_RTPPROXY_OPT_MAX_SESS)) {
		exit_error(PARAMETER_PROBLEM,
		           "RTPPROXY target: addr-a, port-a and max-sess options are required");
	}
}


/* Prints out the targinfo. */
static void
print(const struct ipt_ip *ip,
      const struct ipt_entry_target *target,
      int numeric) {
	const struct xt_rtpproxy_target_info *info
		= (const struct xt_rtpproxy_target_info *)target->data;
	struct in_addr ip_a = { info->gate[0].ip };
	struct in_addr ip_b = { info->gate[1].ip };

	printf("RTPPROXY ");
	printf("gate-a:%s:[%d-%d] ", inet_ntoa(ip_a), info->gate[0].port, info->gate[0].port + 2*info->max_sess);
	printf("gate-b:%s:[%d-%d] ", inet_ntoa(ip_b), info->gate[1].port, info->gate[1].port + 2*info->max_sess);
	printf("max-sess:%d ", info->max_sess);
	printf("expires-timeout:%d ", info->expires_timeout);
	printf("resurrection-timeout:%d", info->resurrection_timeout);
}


static void save(const struct ipt_ip *ip, 
		 const struct ipt_entry_target *target) {
	const struct xt_rtpproxy_target_info *info
		= (const struct xt_rtpproxy_target_info *)target->data;

	if (info->libipt_flags & IPT_RTPPROXY_OPT_ADDR_A) {
		struct in_addr ip = { info->gate[0].ip };
		printf("--addr-a %s ", inet_ntoa(ip));
	}

	if (info->libipt_flags & IPT_RTPPROXY_OPT_ADDR_B) {
		struct in_addr ip = { info->gate[1].ip };
		printf("--addr-b %s ", inet_ntoa(ip));
	}

	if (info->libipt_flags & IPT_RTPPROXY_OPT_PORT_A) {
		printf("--port-a %d ", info->gate[0].port);
	}
	if (info->libipt_flags & IPT_RTPPROXY_OPT_PORT_B) {
		printf("--port-b %d ", info->gate[1].port);
	}

	if (info->libipt_flags & IPT_RTPPROXY_OPT_MAX_SESS) {
		printf("--max-sess %d ", info->max_sess);
	}

	if (info->libipt_flags & IPT_RTPPROXY_OPT_EXPIRES_TIMEOUT) {
		printf("--expires-timeout %d ", info->expires_timeout);
	}
	if (info->libipt_flags & IPT_RTPPROXY_OPT_RESURRECTION_TIMEOUT) {
		printf("--resurrection-timeout %d ", info->resurrection_timeout);
	}
}


static struct iptables_target rtpproxy = { 
	.next		= NULL,
	.name		= "RTPPROXY",
	.version	= IPTABLES_VERSION,
	.size		= IPT_ALIGN(sizeof(struct xt_rtpproxy_target_info)),
	.userspacesize	= offsetof(struct xt_rtpproxy_target_info, expires_timeout),
	.help		= &help,
	.init		= &init,
	.parse		= &parse,
	.final_check	= &final_check,
	.print		= &print,
	.save		= &save,
	.extra_opts	= opts
};

void _init(void) {
	register_target(&rtpproxy);
}

/* libipt RTPPROXY interface functions */
int xt_RTPPROXY_open(struct xt_rtpproxy_handle *handle) {
	handle->sockfd = socket(AF_INET, SOCK_RAW, IPPROTO_RAW);
	if (handle->sockfd < 0) {
		handle->sockfd = -1;
		handle->err_str = "Can't open socket to kernel. Permission denied or ipt_RTPPROXY module not loaded";
		handle->err_no = XT_RTPPROXY_ERR_CANNOT_OPEN_SOCKET;
		return -1;
	}
	return 0;
}

void xt_RTPPROXY_close(struct xt_rtpproxy_handle *handle) {
	close(handle->sockfd);
	handle->sockfd = -1;
}

#define check_version(handle) \
	if ((handle)->info.h.version != XT_RTPPROXY_SOCKOPT_VERSION) { \
		(handle)->err_str = "bad version of response"; \
		(handle)->err_no = XT_RTPPROXY_ERR_BAD_VERSION; \
		return -1; \
	} 

int xt_RTPPROXY_get_switchboards(
	struct xt_rtpproxy_handle *handle,
	struct xt_rtpproxy_sockopt_in_switchboard *in_switchboard_id,
	struct xt_rtpproxy_sockopt_in_sess_id *in_sess_id,
	unsigned int flags,
	struct xt_rtpproxy_switchboard **out_switchboard
) {
	struct xt_rtpproxy_sockopt_header *in_h;
	struct xt_rtpproxy_sockopt_num_session *out_count;
	char buff[
		/* in */
		/*sizeof(struct xt_rtpproxy_sockopt_header)+ */
		sizeof(struct xt_rtpproxy_sockopt_in_switchboard)+
		sizeof(struct xt_rtpproxy_sockopt_in_sess_id)+
		/* out */
		sizeof(struct xt_rtpproxy_sockopt_out_config)+
		sizeof(struct xt_rtpproxy_sockopt_num_session)];
	char *p, *buff2;
	unsigned int sz;
	int res, out_len;

	if (out_switchboard)
		*out_switchboard = NULL;

	if (flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD) {
		/* we must allocate buffer first for switchboard/session data */
		if (!out_switchboard)
			return -1;
		in_h = (void*) buff;
		in_h->version = XT_RTPPROXY_SOCKOPT_VERSION;
		in_h->flags = 0;
		p = &buff[sizeof(struct xt_rtpproxy_sockopt_header)];
		if (in_switchboard_id) {		
			in_h->flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD;
			memcpy(p, in_switchboard_id, sizeof(*in_switchboard_id));
			p+= sizeof(*in_switchboard_id);
		}
		if (in_sess_id) {		
			in_h->flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SESS_ID;
			memcpy(p, in_sess_id, sizeof(*in_sess_id));
			p+= sizeof(*in_sess_id);
		}
		in_h->flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT; 
		sz = sizeof(buff);
		handle->err_no = getsockopt(handle->sockfd, IPPROTO_IP, XT_RTPPROXY_SO_GET_CONFIG, &buff, &sz);
		if (handle->err_no < 0) {
			handle->err_str = "getsockopt(XT_RTPPROXY_SO_GET_CONFIG) error";
			return -1;
		}
		memcpy(&handle->info, buff, sizeof(handle->info));
		check_version(handle);
		p = buff+sizeof(handle->info);
		if ((handle->info.h.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT) == 0) { /* should never happen */
			handle->err_str = "Unexpected error, no session count returned";
			handle->err_no = XT_RTPPROXY_ERR_UNEXPECTED;
			return -1;
		}
		out_count = (void*) p;

		out_len = sizeof(struct xt_rtpproxy_sockopt_out_config);
	/*	if (flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT)
			out_len += sizeof(struct xt_rtpproxy_sockopt_num_session); */
		if (flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD) {
			out_len += out_count->num_switchboard*sizeof(struct xt_rtpproxy_sockopt_switchboard);
			if (flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION)
				out_len += out_count->num_session*sizeof(struct xt_rtpproxy_session);
		}
		buff2 = (void*) malloc(out_len);
		if (!buff2) {
			handle->err_str = "not enough memory";
			handle->err_no = XT_RTPPROXY_ERR_OUT_OF_MEMORY;
			return -1;
		}
		sz = out_len;
	}
	else {
		flags &= ~(XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD|XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION);
		buff2 = buff;	
		sz = sizeof(buff);
	}


	in_h = (void*) buff2;
	in_h->version = XT_RTPPROXY_SOCKOPT_VERSION;
	p = buff2 + sizeof(struct xt_rtpproxy_sockopt_header);
	if (in_switchboard_id) {		
		flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD;
		memcpy(p, in_switchboard_id, sizeof(*in_switchboard_id));
		p+= sizeof(*in_switchboard_id);
	}
	else
		flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD;
	if (in_sess_id) {		
		flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SESS_ID;
		memcpy(p, in_sess_id, sizeof(*in_sess_id));
		p+= sizeof(*in_sess_id);
	}
	else
		flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_IN_SESS_ID;
	flags &= ~XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT;
	in_h->flags = flags;

	handle->err_no = getsockopt(handle->sockfd, IPPROTO_IP, XT_RTPPROXY_SO_GET_CONFIG, buff2, &sz);
	if (handle->err_no < 0) {
		handle->err_str = "getsockopt(XT_RTPPROXY_SO_GET_CONFIG) error";
		res = -1;
		goto exit;
	}
	memcpy(&handle->info, buff2, sizeof(handle->info));
	check_version(handle);
	p = buff2+sizeof(handle->info);

	if (handle->info.h.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT) { /* should never happen */
		p = buff2+sizeof(*out_count);
	}
	if (flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD) {
		/* parse result */
		u_int32_t sw_flags, sess_flags;
		struct xt_rtpproxy_switchboard *sw_prev, *sw;
		struct xt_rtpproxy_session *sess_prev, *sess;
		for (sw_flags = handle->info.h.flags, sw_prev = NULL; sw_flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD; sw_flags = sw->so.flags, sw_prev = sw) {

			sw = malloc(sizeof(*sw));
			if (!sw) {
				handle->err_str = "not enough memory";
				handle->err_no = XT_RTPPROXY_ERR_OUT_OF_MEMORY;
				xt_RTPPROXY_release_switchboards(handle, *out_switchboard);
				res = -1;
				goto exit;
			}
			memset(sw, 0, sizeof(*sw));
			memcpy(&sw->so, p, sizeof(sw->so));
			p+= sizeof(sw->so);
			if (sw_prev) {
				sw->prev = sw_prev;
				sw_prev->next = sw;
			}
			else {
				*out_switchboard = sw;
			}
			for (sess_flags = sw->so.flags, sess_prev = NULL; sess_flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION; sess_flags = sess->flags, sess_prev = sess) {
				sess = malloc(sizeof(*sess));
				if (!sess) {
					handle->err_str = "not enough memory";
					handle->err_no = XT_RTPPROXY_ERR_OUT_OF_MEMORY;
					xt_RTPPROXY_release_switchboards(handle, *out_switchboard);
					res = -1;
					goto exit;
				}
				memset(sess, 0, sizeof(*sess));
				memcpy(sess, p, sizeof(*sess));
				p+= sizeof(*sess);
				sess->next = 0;
				if (sess_prev) {
					sess->prev = sess_prev;
					sess_prev->next = sess;
				}
				else {
					sess->prev = 0;
					sw->sessions = sess;

				}
			}
		}
	}
	res = 0;
exit:
	if (buff != buff2)
		free(buff2);
	return res;
}

void xt_RTPPROXY_release_switchboards(struct xt_rtpproxy_handle *handle, struct xt_rtpproxy_switchboard *switchboard) {
	struct xt_rtpproxy_switchboard *p;
	struct xt_rtpproxy_session *s;
	while (switchboard) {
		while (switchboard->sessions) {
			s = switchboard->sessions;
			switchboard->sessions = switchboard->sessions->next;
			free(s);
		}
		p = switchboard;
		switchboard=switchboard->next;
		free(p);
	}
}


int xt_RTPPROXY_alloc_session(
	struct xt_rtpproxy_handle *handle,
	struct xt_rtpproxy_sockopt_in_switchboard *in_switchboard_id,	/* NULL to alloc at any switchboard */
	struct xt_rtpproxy_sockopt_in_alloc_session *in_session,	/* identify initial state, at least one learning timeout or source ip must be specified! */
	struct xt_rtpproxy_sockopt_switchboard *out_switchboard_id,	/* usefull when in_switchboard_id is NULL, otherwise can be NULL */ 
	struct xt_rtpproxy_session *out_session
) {
	struct xt_rtpproxy_sockopt_header *in_h;
	char buff[
		/* in */
		/*sizeof(struct xt_rtpproxy_sockopt_header)+ */
		sizeof(struct xt_rtpproxy_sockopt_in_switchboard)+
		sizeof(struct xt_rtpproxy_sockopt_in_alloc_session)+
		/* out */
		sizeof(struct xt_rtpproxy_sockopt_out_config)+
		/* sizeof(struct xt_rtpproxy_sockopt_in_switchboard)+ */
		sizeof(struct xt_rtpproxy_session)];
	char *p;
	unsigned int sz, i, j;

	in_h = (void*) buff;
	in_h->version = XT_RTPPROXY_SOCKOPT_VERSION;
	in_h->flags = 0;
	p = &buff[sizeof(struct xt_rtpproxy_sockopt_header)];
	if (in_switchboard_id) {		
		in_h->flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD;
		memcpy(p, in_switchboard_id, sizeof(*in_switchboard_id));
		p+= sizeof(*in_switchboard_id);
	}
	/* test if there is at least one ip or learning timeout */
	for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
		for (j=0; j<2; j++) {
			if ( (in_session->source[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR &&
			      in_session->source[i].stream[j].addr.ip) || 
			     (in_session->source[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT &&
			      in_session->source[i].stream[j].learning_timeout) ) 
				goto ip_test_ok;
		}
	}
	handle->err_str = "at least one ip or learning timeout must be specified";
	handle->err_no = XT_RTPPROXY_ERR_PARAM;
	return -1;

ip_test_ok:		
	memcpy(p, in_session, sizeof(*in_session));
	p+= sizeof(*in_session);
	sz = sizeof(buff);
	handle->err_no = getsockopt(handle->sockfd, IPPROTO_IP, XT_RTPPROXY_SO_GET_ALLOC_SESSION, &buff, &sz);
	if (handle->err_no < 0) {
		handle->err_str = "getsockopt(XT_RTPPROXY_SO_GET_ALLOC_SESSION) error";
		return -1;
	}
	memcpy(&handle->info.h, buff, sizeof(handle->info.h));
	check_version(handle);
	p = buff+sizeof(handle->info.h);
	if (handle->info.h.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_COUNT) { /* should never happen */
		p+= sizeof(struct xt_rtpproxy_sockopt_num_session);
	}
	if (handle->info.h.flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD) {
		if (out_switchboard_id)
			memcpy(out_switchboard_id, p, sizeof(*out_switchboard_id));
		p+= sizeof(*out_switchboard_id);
	}
	else {
		handle->err_str = "Unexpected error, no switchboard returned";
		handle->err_no = XT_RTPPROXY_ERR_UNEXPECTED;
		return -1;
	}
	if (out_switchboard_id->flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION) {
		if (out_session) {
			memcpy(out_session, p, sizeof(*out_session));
			out_session->prev = 0;
			out_session->next = 0;
		}
		p+= sizeof(*out_session);
	}
	else {
		handle->err_str = "Unexpected error, no session returned";
		handle->err_no = XT_RTPPROXY_ERR_UNEXPECTED;
		return -1;
	}
	return 0;
}

int xt_RTPPROXY_update_session(
	struct xt_rtpproxy_handle *handle,
	struct xt_rtpproxy_sockopt_in_switchboard *in_switchboard_id,	/* mandatory to identify updated switchboard */
	struct xt_rtpproxy_sockopt_in_sess_id *in_sess_id,		/* NULL to update all sessions of given switchboard */
	struct xt_rtpproxy_sockopt_in_alloc_session *in_session
) {
	struct xt_rtpproxy_sockopt_header *in_h;
	char buff[sizeof(struct xt_rtpproxy_sockopt_header)+
		sizeof(struct xt_rtpproxy_sockopt_in_switchboard)+
		sizeof(struct xt_rtpproxy_sockopt_in_alloc_session)+
		sizeof(struct xt_rtpproxy_sockopt_in_alloc_session)];
	char *p;

	in_h = (void*) buff;
	in_h->version = XT_RTPPROXY_SOCKOPT_VERSION;
	in_h->flags = 0;
	p = &buff[sizeof(struct xt_rtpproxy_sockopt_header)];
	if (in_switchboard_id) {		
		in_h->flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD;
		memcpy(p, in_switchboard_id, sizeof(*in_switchboard_id));
		p+= sizeof(*in_switchboard_id);
	}
	else {
 		handle->err_str = "switchboard is not identified";
		handle->err_no = XT_RTPPROXY_ERR_PARAM;
		return -1;
	}
	if (in_sess_id) {
		in_h->flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SESS_ID;
		memcpy(p, in_sess_id, sizeof(*in_sess_id));
		p+= sizeof(*in_sess_id);
	}
	memcpy(p, in_session, sizeof(*in_session));
	p+= sizeof(*in_session);
	handle->err_no = setsockopt(handle->sockfd, IPPROTO_IP, XT_RTPPROXY_SO_SET_UPDATE_SESSION, &buff, p-buff);
	if (handle->err_no < 0) {
		handle->err_str = "setsockopt(XT_RTPPROXY_SO_SET_UPDATE_SESSION) error";
		return -1;
	}
	return 0;
}

int xt_RTPPROXY_delete_session(
	struct xt_rtpproxy_handle *handle,
	struct xt_rtpproxy_sockopt_in_switchboard *in_switchboard_id,	/*NULL to delete all switchboards */
	struct xt_rtpproxy_sockopt_in_sess_id *in_sess_id		/* NULL to delete all sessions of given switchboard */
) {
	struct xt_rtpproxy_sockopt_header *in_h;
	char buff[sizeof(struct xt_rtpproxy_sockopt_header)+
		sizeof(struct xt_rtpproxy_sockopt_in_switchboard)+
		sizeof(struct xt_rtpproxy_sockopt_in_alloc_session)];
	char *p;

	in_h = (void*) buff;
	in_h->version = XT_RTPPROXY_SOCKOPT_VERSION;
	in_h->flags = 0;
	p = &buff[sizeof(struct xt_rtpproxy_sockopt_header)];
	if (in_switchboard_id) {		
		in_h->flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SWITCHBOARD;
		memcpy(p, in_switchboard_id, sizeof(*in_switchboard_id));
		p+= sizeof(*in_switchboard_id);
	}
	if (in_sess_id) {
		in_h->flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_SESS_ID;
		memcpy(p, in_sess_id, sizeof(*in_sess_id));
		p+= sizeof(*in_sess_id);
	}
	handle->err_no = setsockopt(handle->sockfd, IPPROTO_IP, XT_RTPPROXY_SO_SET_DELETE_SESSION, &buff, p-buff);
	if (handle->err_no < 0) {
		handle->err_str = "setsockopt(XT_RTPPROXY_SO_SET_DELETE_SESSION) error";
		return -1;
	}
	return 0;
}

