/***************************************************************************
 *   (C) 2007 by Tomas Mandys <tomas.mandys at 2p dot cz>                  *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU Lesser General Public License           *
 *   version 2.1 as published by the Free Software Foundation;             *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#define _GNU_SOURCE
#include <getopt.h>
#include <signal.h>
#include <string.h>
#include <arpa/inet.h>

#include <linux/netfilter/xt_RTPPROXY.h>

#define IPTRTPPROXY_VERSION "O.1"

static int verbose = 0;

#define CMDVAL_ADDR_PORT 0x0001
#define CMDVAL_A_B 0x0002
#define CMDVAL_RTP_RTCP 0x0004

#define CMDVAL_GRP_MASK 0x0F00
#define CMDVAL_GRP_ADDR 0x0100
#define CMDVAL_GRP_RADDR 0x0200
#define CMDVAL_GRP_LEARNING_TIMEOUT 0x0300
#define CMDVAL_GRP_ALWAYS_LEARN 0x0400
#define CMDVAL_GRP_SESSION_ID 0x0500
#define CMDVAL_GRP_LIST 0x0600
#define CMDVAL_GRP_RESET 0x0700

static struct option long_options[] = {
	{ "addr-a",			1, 0, CMDVAL_GRP_ADDR },
	{ "addr-b",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_A_B },
	{ "port-a",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_ADDR_PORT },
	{ "port-b",			1, 0, CMDVAL_GRP_ADDR|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtp-addr-a",			1, 0, CMDVAL_GRP_RADDR },
	{ "rtcp-addr-a",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP },
	{ "rtp-port-a",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_ADDR_PORT },
	{ "rtcp-port-a",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_ADDR_PORT },
	{ "rtp-addr-b",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_A_B },
	{ "rtcp-addr-b",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_A_B },
	{ "rtp-port-b",			1, 0, CMDVAL_GRP_RADDR|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtcp-port-b",		1, 0, CMDVAL_GRP_RADDR|CMDVAL_RTP_RTCP|CMDVAL_ADDR_PORT|CMDVAL_A_B },
	{ "rtp-learning-timeout-a",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT },
	{ "rtp-learning-timeout-b",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_A_B },
	{ "rtcp-learning-timeout-a",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_RTP_RTCP },
	{ "rtcp-learning-timeout-b",	1, 0, CMDVAL_GRP_LEARNING_TIMEOUT|CMDVAL_RTP_RTCP|CMDVAL_A_B },
	{ "always-learn-a",		0, 0, CMDVAL_GRP_ALWAYS_LEARN },
	{ "always-learn-b",		0, 0, CMDVAL_GRP_ALWAYS_LEARN|CMDVAL_A_B },

	{ "sess-id-lo",			1, 0, CMDVAL_GRP_SESSION_ID },
	{ "sess-id-hi",			1, 0, CMDVAL_GRP_SESSION_ID|1 },

	{ "no-switchboard",		0, 0, CMDVAL_GRP_LIST|1 },
	{ "no-session",			0, 0, CMDVAL_GRP_LIST|2 },

	{ "reset-global-stat",		0, 0, CMDVAL_GRP_RESET|0 },
	{ "reset-switchboard-stat",	0, 0, CMDVAL_GRP_RESET|1 },
	{ "reset-packet-stat",		0, 0, CMDVAL_GRP_RESET|2 },
	{ "force-switchboard-audit",	0, 0, CMDVAL_GRP_RESET|3 },

	{ "verbose",			0, 0, 'v' },
	{ "help",			0, 0, 'h' },
	{ 0 }
};


struct enum_entry {
        int id;
        char* name;
        char* descr;
};

enum {cmdNone=0, cmdInfo, cmdList, cmdAlloc, cmdUpdate, cmdDelete};

static struct enum_entry enum_command[] = {
	{id: cmdInfo, 	name: "info", descr: 
		"get RTP proxy global information"
	},
	{id: cmdList, 	name: "list",	descr: 
		"list RTP sessions of specified switchboard. If no switchboard specified then all switchboards are listed. "
		"Session-id may specify session range. "
		"Specify verbosity via --list-xxx options."
	},
	{id: cmdAlloc,	name: "alloc",	descr: 
		"alloc new RTP session. If switchboard is not specified then session is allocated at any switchboard. "
		"At least one remote client addr/port specification is mandatory."
	},
	{id: cmdUpdate,	name: "update",	descr: 
		"update RTP session(s) at specified switchboard. If session-id range is not specified then all sessions of particular switchboard are updated."
	},
	{id: cmdDelete,	name: "delete",	descr: 
		"delete RTP session(s). If switchboard is not specified then all switchboard sessions are deleted according session-id. "
		"If session-id is not specified then all sessions are deleted."
	},

	{id: cmdNone, name: NULL}
};

static struct enum_entry *enum_by_id(int id, struct enum_entry *e) {
	int i;
	for (i=0; e[i].name != NULL; i++) {
		if (e[i].id == id) 
			return e+i;
	}
	return NULL;
}

static struct enum_entry *enum_by_name(char *name, struct enum_entry *e) {
	int i;
	for (i=0; e[i].name != NULL; i++) {
		if (strcmp(name, e[i].name)==0) 
			return e+i;
	}
	return NULL;
}

static void print_opt (int cmd) {
	struct option *o;
	int i;
	for (o=long_options; o->name; o++) {
		if ((o->val & CMDVAL_GRP_MASK) == cmd) {
			printf("  --%s ", o->name);
			switch (cmd) {
				case CMDVAL_GRP_ADDR:
					printf("<%s  RTP proxy gate-%c %s", 
						(o->val & CMDVAL_ADDR_PORT)?"port>":"ip>  ",
						'a'+((o->val & CMDVAL_A_B) != 0),
						(o->val & CMDVAL_ADDR_PORT)?"port":"ip"
					);
					break;

				case CMDVAL_GRP_RADDR:
					printf("<%s  %s client-%c %s", 
						(o->val & CMDVAL_ADDR_PORT)?"port>":"ip>  ",
						(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
						'a'+((o->val & CMDVAL_A_B) != 0),
						(o->val & CMDVAL_ADDR_PORT)?"port":"ip"
					);
					break;

				case CMDVAL_GRP_LEARNING_TIMEOUT:
					printf("<msec>  %s client-%c learning timeout", 
						(o->val & CMDVAL_RTP_RTCP)?"RTCP":" RTP",
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;

				case CMDVAL_GRP_ALWAYS_LEARN:
					printf("     always learn RTP client-%c", 
						'a'+((o->val & CMDVAL_A_B) != 0)
					);
					break;

				case CMDVAL_GRP_SESSION_ID:
					printf("<id>  session-id %s limit", 
						(o->val & 0x0F)?"hi":"lo"
					);
					break;

				case CMDVAL_GRP_LIST:
					for (i=strlen(o->name); i<16; i++) printf(" ");
					switch (o->val & 0x0F) {
						case 1:
							printf("  don't list switchboards and sessions");
							break; 
						case 2:
							printf("  don't list sessions");
							break; 
						default: ;
					}
					break;

				case CMDVAL_GRP_RESET:
					for (i=strlen(o->name); i<23; i++) printf(" ");
					switch (o->val & 0x0F) {
						case 0:
							printf("  reset global statistics");
							break; 
						case 1:
							printf("  reset switchboard statistics");
							break; 
						case 2:
							printf("  reset switchboard packet statistics");
							break; 
						case 3:
							printf("  force switchboard audit");
							break; 
						default: ;
					}
					break;

				default: ;
			}
			printf("\n");
		}
	}
}

static print_version() {
	printf("iptables RTPPROXY helper, v%s\n", IPTRTPPROXY_VERSION);
}

static help(int cmd, char *name) {
	struct enum_entry *e;
	print_version();
	if (!cmd) {
		printf("Usage: %s [options] <command> [options]\n", basename(name));
		printf("Type '%s -h <command>' for help on a specific command.\n", basename(name));

		printf("\n");
		printf("Command list:\n");
		for (e=enum_command; e->name; e++) {
			printf("  %-8s  %s\n", e->name, e->descr);
		}
	}
	else {
		e = enum_by_id(cmd, enum_command);
		printf("Usage: %s %s [options]\n", basename(name), e->name);

		printf("%s: %s\n", e->name, e->descr);

		if (cmd != cmdInfo) {
			printf( "\n"
				"Switchboard identification:\n");
			print_opt(CMDVAL_GRP_ADDR);
	
			if (cmd != cmdAlloc) {
				printf( "\n"
					"Session range:\n");
				print_opt(CMDVAL_GRP_SESSION_ID);
			}
	
			if (cmd == cmdAlloc || cmd == cmdUpdate) {
				printf( "\n"
					"RTP client parameters:\n");
				print_opt(CMDVAL_GRP_RADDR);
				print_opt(CMDVAL_GRP_LEARNING_TIMEOUT);
				print_opt(CMDVAL_GRP_ALWAYS_LEARN);
			}

			if (cmd == cmdList) {
				printf( "\n"
					"List verbosity:\n");
				print_opt(CMDVAL_GRP_LIST);
			}
	
		}
		printf( "\n"
			"Other options:\n");
		print_opt(CMDVAL_GRP_RESET);
	}

	printf("\n");
	printf("Common options:\n");
	printf("  -h,--help      print this help\n");
	printf("  -v,--verbose   print more info, mostly to stderr\n");
	printf("  -V             print version\n");
	printf("\n");
}

static char *ip2s(u_int32_t ip) {
	struct in_addr ip2 = { ip };
	return inet_ntoa(ip2);
} 
/*static char *ip2s(u_int32_t addr) {
    static char buf[17];
    const unsigned char *bytep;

    bytep = (const unsigned char *) &addr;
    snprintf(buf, 16, "%u.%u.%u.%u", bytep[0], bytep[1], bytep[2], bytep[3]);
    buf[16] = 0;
    return buf;
}
*/

static void print_header(struct xt_rtpproxy_handle *handle) {
	printf("header: {version: 0x%x, flags: 0x%x}\n", handle->info.h.version, handle->info.h.flags);
}

static void print_stamp(u_int32_t t) {
	time_t tt;
	char *c, *c2;	
	tt = (time_t) t;
	c = ctime(&tt);
	for (c2 = c+strlen(c); c2 >= c && *c2 <= ' ' ; c2--) {
		*c2 = '\0';
	}
	printf("{tick: %u, time: %s}", t, c);
};

static void print_info(struct xt_rtpproxy_handle *handle) {
	static char *cmd_s[XT_RTPPROXY_MAX_CMD] = {"update", "delete", "config", "alloc"};
	int i;
	if (verbose) print_header(handle);

	printf("started: "); print_stamp(handle->info.started); printf("\n");
	printf("total: {switchboards: %d, sessions: %d}\n", handle->info.total.num_switchboard, handle->info.total.num_session); 
	printf("stat: {ok: %d, fail: %d, perm: %d, unknown: %d", 
		handle->info.stat.ok, 
		handle->info.stat.fail, 
		handle->info.stat.perm, 
		handle->info.stat.unknown
	);
	for (i=0; i<XT_RTPPROXY_MAX_CMD; i++) {
		printf(", %s: %d", cmd_s[i], handle->info.stat.cmd[i]); 
	}
	printf("}\n");
}

static void print_stat(struct xt_rtpproxy_stat *stat) {
	printf("packets: {in: %d, out: %d}, bytes: {in: %d, out: %d}",
		stat->packets_in,
		stat->packets_out,
		stat->bytes_in,
		stat->bytes_out
	);
};

static void print_in_switchboard(struct xt_rtpproxy_sockopt_in_switchboard *s) {
	fprintf(stderr, "  switchboard: ");
	if (s) {
		fprintf(stderr, "{gate-a: {ip: %s, port: %d}, ", 
			ip2s(s->gate[0].ip), 
			s->gate[0].port
		); /* ip2s ... result in static buffer */
		fprintf(stderr, "gate-b: {ip: %s, port: %d}}", 
			ip2s(s->gate[1].ip), 
			s->gate[1].port
		); /* ip2s ... result in static buffer */
	}
	else {
		fprintf(stderr, "NULL");
	}
}

static void print_in_session(struct xt_rtpproxy_sockopt_in_alloc_session *s) {
	int i, j;
	fprintf(stderr, "  session: ");
	if (s) {
		fprintf(stderr, "{\n");
		for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
			fprintf(stderr, "    gate-%c: {\n", 'a'+i);
			for (j=0; j<2; j++) {
				if (s->source[i].stream[j].flags != 0) {
					fprintf(stderr, "      %s: {", j?"rtcp":"rtp");
					if (s->source[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR) {
						fprintf(stderr, "addr: {ip: %s, port: %d}", ip2s(s->source[i].stream[j].addr.ip), s->source[i].stream[j].addr.port);
					}
					if (s->source[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT) {
						if (s->source[i].stream[j].flags & XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR)
							fprintf(stderr, ", ");
						fprintf(stderr, "learn-timeout: %d", s->source[i].stream[j].learning_timeout);
					}
					fprintf(stderr, "},\n");
				}
			}
			fprintf(stderr, "      always-learn: %d\n", s->source[i].always_learn);
			fprintf(stderr, "    }%s\n", (i<XT_RTPPROXY_MAX_GATE-1)?",":"");
		}
		fprintf(stderr, "  }");
	}
	else {
		fprintf(stderr, "NULL");
	}
}

static void print_in_session_id(struct xt_rtpproxy_sockopt_in_sess_id *s) {
	fprintf(stderr, "  sess-id: ");
	if (s) {
		fprintf(stderr, "{min: %d, max: %d}", s->sess_id_min, s->sess_id_max);
	}
	else {
		fprintf(stderr, "NULL");
	}
}

static char stop_flag = 0;
static void sig_term(int signr) {
    signal(SIGINT, SIG_IGN);
    signal(SIGQUIT, SIG_IGN);
    signal(SIGTERM, SIG_IGN);
    stop_flag = 1;
}
#define SIGNAL(sig) \
	if (signal(sig, sig_term) == SIG_ERR) {\
		printf("can't install signal handler for %s\n", #sig);\
	exit (-1);\
}
 
static u_int32_t s2ip(char *optarg) {
	u_int32_t res = 0;
	if (!inet_aton(optarg, (struct in_addr*) &res)) {
		fprintf(stderr, "Invalid IP address %s\n", optarg);
		exit(-1);
	}
	return res;
}

static u_int32_t s2u(char *optarg) {
	long res = atol(optarg);
	if (res < 0) {
		fprintf(stderr, "Invalid number %s\n", optarg);
		exit(-1);
	}
	return res;
}

int main(int argc, char** argv) { 

	int cmd = cmdNone, help_flag = 0;
	static char short_opts[100] = "+HvV";
	struct xt_rtpproxy_sockopt_in_alloc_session in_session;
	struct xt_rtpproxy_sockopt_in_switchboard in_switchboard;
	struct xt_rtpproxy_sockopt_in_sess_id in_sess_id;
	struct xt_rtpproxy_handle handle;
	int i, j, in_switchboard_fl, in_sess_id_fl;
	u_int32_t in_flags, list_verbose;

	if (argc <= 1) {
		help(cmd, argv[0]);
		return 2;
	}

	memset(&in_session, 0, sizeof(in_session));
	memset(&in_switchboard, 0, sizeof(in_switchboard));
	memset(&in_sess_id, 0xFF, sizeof(in_sess_id));

	j = strlen(short_opts);
	for (i=0; long_options[i].name; i++) {
		if (long_options[i].val > ' ' && long_options[i].val < 128) {
			short_opts[j++] = (char) long_options[i].val;
			if (long_options[i].has_arg != no_argument) 
				short_opts[j++] = ':';
			if (long_options[i].has_arg == optional_argument) 
				short_opts[j++] = ':';
			short_opts[j] = '\0';
		}
	}

	in_flags = 0;
	list_verbose = 0;
	while (optind < argc) {
		int option_index;
		int c;
		option_index = -1;
		c = getopt_long_only (argc, argv, short_opts, long_options, &option_index);
		if (c == -1) {
			if (optind >= argc) 
				break;

			if (!cmd) {
				struct enum_entry *e;
				e = enum_by_name(argv[optind], enum_command);
				if (!e) {
					fprintf(stderr, "ERROR: Unknown command: %s\n", argv[optind]);
					goto destroy;
				}
				cmd = e->id;
				optind++;
			} 
			else {
				fprintf(stderr, "ERROR: Multiple commands not allowed (%s)\n", argv[optind]);
				return 2;
			}
		}
		else {
			
			switch ((c & CMDVAL_GRP_MASK)?c & CMDVAL_GRP_MASK:c) {
				case CMDVAL_GRP_ADDR:
					if ((c & CMDVAL_ADDR_PORT) != 0)
						in_switchboard.gate[(c & CMDVAL_A_B) != 0].port = s2u(optarg);
					else
						in_switchboard.gate[(c & CMDVAL_A_B) != 0].ip = s2ip(optarg);
					break;
				case CMDVAL_GRP_RADDR:
					if ((c & CMDVAL_ADDR_PORT) != 0)
						in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].addr.port = s2u(optarg);
					else
						in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].addr.ip = s2ip(optarg);
					in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_ADDR;
					break;
				case CMDVAL_GRP_LEARNING_TIMEOUT:
					in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].learning_timeout = s2u(optarg);
					in_session.source[(c & CMDVAL_A_B) != 0].stream[(c & CMDVAL_RTP_RTCP) != 0].flags |= XT_RTPPROXY_SOCKOPT_FLAG_SESSION_LEARNING_TIMEOUT;
					break;
				case CMDVAL_GRP_ALWAYS_LEARN:
					in_session.source[(c & CMDVAL_A_B) != 0].always_learn = 1;
					break;
				case CMDVAL_GRP_SESSION_ID:
					if (c & 0x01)
						in_sess_id.sess_id_max = s2u(optarg);
					else
						in_sess_id.sess_id_min = s2u(optarg);
					break;
				case CMDVAL_GRP_LIST:
					switch (c & 0x0F) {
						case 1:
							list_verbose |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD | XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
							break;
						case 2:
							list_verbose |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
							break;
						default: ;
					}
					break;

				case CMDVAL_GRP_RESET:
					switch (c & 0x0F) {
						case 0:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_GLOBAL_STAT;
							break;
						case 1:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_SWITCHBOARD_STAT;
							break;
						case 2:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_RESET_SWITCHBOARD_PACKET_STAT;
							break;
						case 3:
							in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_IN_FORCE_SWITCHBOARD_AUDIT;
							break;
						default: ;
					}
					break;
				case 'v':
					verbose++;
					break;
				case 'V':
					print_version();
					return 0;
				case 'h':
				case 'H':
					help_flag = 1;
					break;
				case '?':
					break;
				default:
					fprintf(stderr, "getopt_long_only returned character code 0x%x\n", c);
					break;
			}
		}
		if (cmd && help_flag)
			break;
	}
	if (!cmd || help_flag) {
		help(cmd, argv[0]);
		return help_flag?0:2;
	}

	SIGNAL(SIGTERM);
	SIGNAL(SIGINT);
	SIGNAL(SIGQUIT);

	/* switchboard may be optional but when at least one port or ip is defined then complete definition is mandatory */
	in_switchboard_fl = 0;
	for (i=0; i<XT_RTPPROXY_MAX_GATE; i++) {
		if (in_switchboard.gate[i].ip || in_switchboard.gate[i].port)
			in_switchboard_fl = 1;
		if (i>0) {
			if (!in_switchboard.gate[i].ip)
				in_switchboard.gate[i].ip = in_switchboard.gate[0].ip;
			if (!in_switchboard.gate[i].port)
				in_switchboard.gate[i].port = in_switchboard.gate[0].port;
		}
		if (in_switchboard_fl && (!in_switchboard.gate[i].ip || !in_switchboard.gate[i].port)) {
			fprintf(stderr, "ERROR: uncomplete switchboard identification\n");
			return 2;
		} 
	}

	if (in_sess_id.sess_id_max == 0xFFFF)
		in_sess_id.sess_id_max = in_sess_id.sess_id_min;
	if (in_sess_id.sess_id_min == 0xFFFF)
		in_sess_id.sess_id_min = in_sess_id.sess_id_max;
	if (in_sess_id.sess_id_max < in_sess_id.sess_id_min) {
		u_int16_t x = in_sess_id.sess_id_max;
		in_sess_id.sess_id_max = in_sess_id.sess_id_min;
		in_sess_id.sess_id_min = x;
	}
	in_sess_id_fl = in_sess_id.sess_id_min != 0xFFFF && in_sess_id.sess_id_max != 0xFFFF;
	
	if (verbose) fprintf(stderr, "xt_RTPPROXY_open()\n");
	if (xt_RTPPROXY_open(&handle) < 0) goto err;

	switch (cmd) {
		case cmdInfo:
			if (verbose) fprintf(stderr, "xt_RTPPROXY_get_info(flags:0x%x)\n", in_flags);
			if (xt_RTPPROXY_get_info(&handle, in_flags) < 0)
				goto err;
			print_info(&handle);
			break;

		case cmdList: {
			struct xt_rtpproxy_switchboard *out_switchboard, *sb;
			static char* state_s[] = {"none", "init1", "init2", "forward1", "forward2", "destroyed", "expired"};
			in_flags |= XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD | XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION;
			in_flags &= ~list_verbose;  
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_get_switchboards(\n");
				print_in_switchboard(in_switchboard_fl?&in_switchboard:NULL); fprintf(stderr, ",\n");
				print_in_session_id(in_sess_id_fl?&in_sess_id:NULL); fprintf(stderr, ",\n");
				fprintf(stderr, "  flags: 0x%x\n", in_flags); 
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_get_switchboards(&handle, in_switchboard_fl?&in_switchboard:NULL, in_sess_id_fl?&in_sess_id:NULL, in_flags, &out_switchboard) < 0)
				goto err;
			if (verbose) print_info(&handle);

			if (in_flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SWITCHBOARD) {
				printf("{\n");
				for (sb=out_switchboard; sb; sb=sb->next) {
					int j;
					printf("  {\n");
					printf("    switchboard: {\n");
					printf("      gate: {\n");
					printf("        a: {ip: %s, port: %d}, stat: ",
						ip2s(sb->so.gate[0].addr.ip), 
						sb->so.gate[0].addr.port
					); /* ip2s ... result in static buffer */
					print_stat(&sb->so.gate[0].stat);
					printf("\n");
					printf("        b: {ip: %s, port: %d}, stat: ", 
						ip2s(sb->so.gate[1].addr.ip),
						sb->so.gate[1].addr.port
					); /* ip2s ... result in static buffer */
					print_stat(&sb->so.gate[1].stat);
					printf("\n      }\n");
					printf("      max-sess: %d, expires-timeout: %d, resurrection-timeout: %d\n", 
						sb->so.max_sess, 
						sb->so.expires_timeout, 
						sb->so.resurrection_timeout
					);
					printf("      stat: {\n");
					printf("        state: {");
					for (j=0; j<xt_rtpproxy_MAX_STATES; j++) {
						if (j) printf(", ");
						printf("%s: %d", state_s[j], sb->so.stat.state[j]);
					}
					printf("}\n");
					printf("        counter: {alloc: %d, alloc_failed: %d, updated: %d, destroyed: %d, expired: %d}\n",
						sb->so.stat.counter.alloc,
						sb->so.stat.counter.alloc_failed,
						sb->so.stat.counter.updated,
						sb->so.stat.counter.destroyed,
						sb->so.stat.counter.expired 
					);
					printf("      }\n");
					printf("    }\n");
					if (in_flags & XT_RTPPROXY_SOCKOPT_FLAG_OUT_SESSION) {
						struct xt_rtpproxy_session *ss;
						printf("    sessions: {\n");
						for (ss=sb->sessions; ss; ss=ss->next) {
							int g;
							printf("      {\n");
							printf("        state: %s, sess-id: %d, expires: ", state_s[ss->state], ss->sess_id);
							print_stamp(ss->expires);
							printf("\n");
							printf("        gate: {\n");
							for (g=0; g<XT_RTPPROXY_MAX_GATE; g++) {
								int st;
								printf("        %c: {\n", 'a'+g);
								for (st=0; st<2; st++) {
									printf("            %s: {port: %d, source: {ip: %s, port: %d}, learning:  \n", 
										st==0?"rtp":"rtcp",
										ss->gate[g].stream[st].port,
										ip2s(ss->gate[g].stream[st].source.ip),
										ss->gate[g].stream[st].source.port
									);
									print_stamp(ss->gate[g].stream[st].learning),
									printf(", last_packet: ");
									print_stamp(ss->gate[g].stream[st].last_packet_stamp);
									printf("}\n");
								}
								printf("            always_learn: %d\n", ss->gate[g].always_learn);
								printf("            state: "); print_stat(&ss->gate[g].stat); printf("\n");
								
								printf("          }\n");
							}
							printf("        }\n");
							printf("      }\n");
						}
						printf("    }\n");
					}
					printf("  }\n");
				}
				printf("}\n");
			}
			xt_RTPPROXY_release_switchboards(&handle, out_switchboard);
			break;
		}

		case cmdAlloc: {
			struct xt_rtpproxy_sockopt_switchboard out_switchboard_id;
			struct xt_rtpproxy_session out_session;
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_alloc_session(\n");
				print_in_switchboard(in_switchboard_fl?&in_switchboard:NULL); fprintf(stderr, ",\n");
				print_in_session(&in_session); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_alloc_session(&handle, in_switchboard_fl?&in_switchboard:NULL, &in_session, &out_switchboard_id, &out_session) < 0)
				goto err; 
			if (verbose) print_header(&handle);

			printf("switchboard: {gate-a: {ip: %s, port: %d}, ", 
				ip2s(out_switchboard_id.gate[0].addr.ip), 
				out_switchboard_id.gate[0].addr.port
			); /* ip2s ... result in static buffer */
			printf("gate-b: {ip: %s, port: %d}},\n", 
				ip2s(out_switchboard_id.gate[1].addr.ip), 
				out_switchboard_id.gate[1].addr.port
			); /* ip2s ... result in static buffer */
			printf("session: {sess-id: %d, port: {a: {rtp: %d, rtcp: %d}, b: {rtp: %d, rtcp: %d}}}\n",
				out_session.sess_id,
				out_session.gate[0].stream[0].port,
				out_session.gate[0].stream[1].port,
				out_session.gate[1].stream[0].port,
				out_session.gate[1].stream[1].port
			);
			break;
		}
		case cmdUpdate:
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_update_session(\n");
				print_in_switchboard(in_switchboard_fl?&in_switchboard:NULL); fprintf(stderr, ",\n");
				print_in_session_id(in_sess_id_fl?&in_sess_id:NULL); fprintf(stderr, ",\n");
				print_in_session(&in_session); fprintf(stderr, "\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_update_session(&handle, in_switchboard_fl?&in_switchboard:NULL, in_sess_id_fl?&in_sess_id:NULL, &in_session) < 0)
				goto err;
			if (verbose) print_header(&handle);
			break;

		case cmdDelete:
			if (verbose) {
				fprintf(stderr, "xt_RTPPROXY_delete_session(\n");
				print_in_switchboard(in_switchboard_fl?&in_switchboard:NULL); fprintf(stderr, ",\n");
				print_in_session_id(in_sess_id_fl?&in_sess_id:NULL); fprintf(stderr, ",\n");
				fprintf(stderr, ")\n");
			}
			if (xt_RTPPROXY_delete_session(&handle, in_switchboard_fl?&in_switchboard:NULL, in_sess_id_fl?&in_sess_id:NULL) < 0)
				goto err; 
			if (verbose) print_header(&handle);

			break;
	}

destroy:
	if (verbose) fprintf(stderr, "xt_RTPPROXY_close()\n");
	xt_RTPPROXY_close(&handle);
	if (verbose) fprintf(stderr, "Done\n");
	return 0;

err:
	fprintf(stderr, "ERROR: %s (%d)\n", handle.err_str, handle.err_no);
	if (handle.sockfd >= 0) {
		if (verbose) fprintf(stderr, "xt_RTPPROXY_close()\n");
		xt_RTPPROXY_close(&handle);
	}
	if (verbose) fprintf(stderr, "Done\n");
	return 1;
}
